# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

# astropy/coordinates/angle_parsetab.py
# This file is automatically generated. Do not edit.
_tabversion = '3.2'

_lr_method = 'LALR'

_lr_signature = b"'\xecW6\xc1\x00\x86\x9f\x84\x0er\xa4\xfb\x91\xa6;"

_lr_action_items = {'HOUR':([2,8,10,17,18,21,25,26,27,28,34,],[-15,12,-14,20,-16,-12,-13,-9,-8,-10,-11,]),'DEGREE':([2,8,10,17,18,21,25,26,27,28,34,],[-15,13,-14,19,-16,-12,-13,-9,-8,-10,-11,]),'SIGN':([0,],[5,]),'SECOND':([2,8,10,17,18,21,25,26,27,28,32,33,34,],[-15,14,-14,-17,-16,-12,-13,-9,-8,-10,35,36,-11,]),'COLON':([17,28,],[22,31,]),'UINT':([0,5,11,17,19,20,21,22,29,30,31,],[-7,-6,17,21,23,24,26,28,26,26,26,]),'SIMPLE_UNIT':([2,8,10,17,18,21,25,26,27,28,34,],[-15,15,-14,-17,-16,-12,-13,-9,-8,-10,-11,]),'UFLOAT':([0,5,11,21,29,30,31,],[-7,-6,18,27,27,27,27,]),'MINUTE':([2,8,10,17,18,21,23,24,25,26,27,28,34,],[-15,16,-14,-17,-16,-12,29,30,-13,-9,-8,-10,-11,]),'$end':([1,2,3,4,6,7,8,9,10,12,13,14,15,16,17,18,19,20,21,23,24,25,26,27,28,29,30,32,33,34,35,36,],[-4,-15,0,-5,-3,-1,-30,-2,-14,-23,-29,-32,-31,-33,-17,-16,-24,-18,-12,-25,-19,-13,-9,-8,-10,-26,-20,-27,-21,-11,-28,-22,]),}

_lr_action = { }
for _k, _v in _lr_action_items.items():
    for _x,_y in zip(_v[0],_v[1]):
        if not _x in _lr_action:  _lr_action[_x] = { }
        _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'arcminute':([0,],[1,]),'angle':([0,],[3,]),'simple':([0,],[4,]),'arcsecond':([0,],[6,]),'hms':([0,],[7,]),'generic':([0,],[8,]),'dms':([0,],[9,]),'colon':([0,],[10,]),'spaced':([0,],[2,]),'sign':([0,],[11,]),'ufloat':([21,29,30,31,],[25,32,33,34,]),}

_lr_goto = { }
for _k, _v in _lr_goto_items.items():
    for _x,_y in zip(_v[0],_v[1]):
        if not _x in _lr_goto: _lr_goto[_x] = { }
        _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> angle","S'",1,None,None,None),
  ('angle -> hms','angle',1,'p_angle','astropy/coordinates/angle_utilities.py',115),
  ('angle -> dms','angle',1,'p_angle','astropy/coordinates/angle_utilities.py',116),
  ('angle -> arcsecond','angle',1,'p_angle','astropy/coordinates/angle_utilities.py',117),
  ('angle -> arcminute','angle',1,'p_angle','astropy/coordinates/angle_utilities.py',118),
  ('angle -> simple','angle',1,'p_angle','astropy/coordinates/angle_utilities.py',119),
  ('sign -> SIGN','sign',1,'p_sign','astropy/coordinates/angle_utilities.py',125),
  ('sign -> <empty>','sign',0,'p_sign','astropy/coordinates/angle_utilities.py',126),
  ('ufloat -> UFLOAT','ufloat',1,'p_ufloat','astropy/coordinates/angle_utilities.py',135),
  ('ufloat -> UINT','ufloat',1,'p_ufloat','astropy/coordinates/angle_utilities.py',136),
  ('colon -> sign UINT COLON UINT','colon',4,'p_colon','astropy/coordinates/angle_utilities.py',142),
  ('colon -> sign UINT COLON UINT COLON ufloat','colon',6,'p_colon','astropy/coordinates/angle_utilities.py',143),
  ('spaced -> sign UINT UINT','spaced',3,'p_spaced','astropy/coordinates/angle_utilities.py',152),
  ('spaced -> sign UINT UINT ufloat','spaced',4,'p_spaced','astropy/coordinates/angle_utilities.py',153),
  ('generic -> colon','generic',1,'p_generic','astropy/coordinates/angle_utilities.py',162),
  ('generic -> spaced','generic',1,'p_generic','astropy/coordinates/angle_utilities.py',163),
  ('generic -> sign UFLOAT','generic',2,'p_generic','astropy/coordinates/angle_utilities.py',164),
  ('generic -> sign UINT','generic',2,'p_generic','astropy/coordinates/angle_utilities.py',165),
  ('hms -> sign UINT HOUR','hms',3,'p_hms','astropy/coordinates/angle_utilities.py',174),
  ('hms -> sign UINT HOUR UINT','hms',4,'p_hms','astropy/coordinates/angle_utilities.py',175),
  ('hms -> sign UINT HOUR UINT MINUTE','hms',5,'p_hms','astropy/coordinates/angle_utilities.py',176),
  ('hms -> sign UINT HOUR UINT MINUTE ufloat','hms',6,'p_hms','astropy/coordinates/angle_utilities.py',177),
  ('hms -> sign UINT HOUR UINT MINUTE ufloat SECOND','hms',7,'p_hms','astropy/coordinates/angle_utilities.py',178),
  ('hms -> generic HOUR','hms',2,'p_hms','astropy/coordinates/angle_utilities.py',179),
  ('dms -> sign UINT DEGREE','dms',3,'p_dms','astropy/coordinates/angle_utilities.py',192),
  ('dms -> sign UINT DEGREE UINT','dms',4,'p_dms','astropy/coordinates/angle_utilities.py',193),
  ('dms -> sign UINT DEGREE UINT MINUTE','dms',5,'p_dms','astropy/coordinates/angle_utilities.py',194),
  ('dms -> sign UINT DEGREE UINT MINUTE ufloat','dms',6,'p_dms','astropy/coordinates/angle_utilities.py',195),
  ('dms -> sign UINT DEGREE UINT MINUTE ufloat SECOND','dms',7,'p_dms','astropy/coordinates/angle_utilities.py',196),
  ('dms -> generic DEGREE','dms',2,'p_dms','astropy/coordinates/angle_utilities.py',197),
  ('simple -> generic','simple',1,'p_simple','astropy/coordinates/angle_utilities.py',210),
  ('simple -> generic SIMPLE_UNIT','simple',2,'p_simple','astropy/coordinates/angle_utilities.py',211),
  ('arcsecond -> generic SECOND','arcsecond',2,'p_arcsecond','astropy/coordinates/angle_utilities.py',220),
  ('arcminute -> generic MINUTE','arcminute',2,'p_arcminute','astropy/coordinates/angle_utilities.py',226),
]
