# **Please Note**: ``astropy.sphinx`` exists only for backward-compatibility
# purposes - it has now been moved to the separate astropy-helpers package,
# located at https://github.com/astropy/astropy-helpers. Any new development or
# bug fixes should be done there.
"""
This package contains utilities and extensions for the Astropy sphinx
documentation.  In particular, the `astropy.sphinx.conf` should be imported by
the sphinx ``conf.py`` file for affiliated packages that wish to make use of
the Astropy documentation format.
"""
